@@ -21,6 +21,12 @@
 #include "apc_iterator.h"
 #include "apc_cache.h"
 #include "apc_strings.h"
+#include "apc_time.h"
+#if PHP_VERSION_ID >= 80000
+# include "apc_iterator_arginfo.h"
+#else
+# include "apc_iterator_legacy_arginfo.h"
+#endif
 
 #include "ext/standard/md5.h"
 #include "SAPI.h"
@@ -33,6 +39,12 @@ zend_class_entry* apc_iterator_get_ce(void) {
 	return apc_iterator_ce;
 }
 
+#define ENSURE_INITIALIZED(iterator) \
+	if (!(iterator)->initialized) { \
+		zend_throw_error(NULL, "Trying to use uninitialized APCUIterator"); \
+		return; \
+	}
+
 /* {{{ apc_iterator_item */
 static apc_iterator_item_t* apc_iterator_item_ctor(
 		apc_iterator_t *iterator, apc_cache_entry_t *entry) {
@@ -43,7 +55,7 @@ static apc_iterator_item_t* apc_iterator_item_ctor(
 	array_init(&item->value);
 	ht = Z_ARRVAL(item->value);
 
-	item->key = zend_string_dup(entry->key, 0);
+	item->key = zend_string_dup(&entry->key, 0);
 
 	if (APC_ITER_TYPE & iterator->format) {
 		ZVAL_STR_COPY(&zv, apc_str_user);
@@ -98,13 +110,6 @@ static apc_iterator_item_t* apc_iterator_item_ctor(
 }
 /* }}} */
 
-/* {{{ apc_iterator_clone */
-static zend_object* apc_iterator_clone(zval *zobject) {
-	apc_error(APC_ITERATOR_NAME " object cannot be cloned");
-	return NULL;
-}
-/* }}} */
-
 /* {{{ apc_iterator_item_dtor */
 static void apc_iterator_item_dtor(apc_iterator_item_t *item) {
 	zend_string_release(item->key);
@@ -128,14 +133,12 @@ static void apc_iterator_free(zend_object *object) {
 
 	apc_stack_destroy(iterator->stack);
 
-#ifdef ITERATOR_PCRE
 	if (iterator->regex) {
 		zend_string_release(iterator->regex);
-# if PHP_VERSION_ID >= 70300
+#if PHP_VERSION_ID >= 70300
 		pcre2_match_data_free(iterator->re_match_data);
-# endif
-	}
 #endif
+	}
 
 	if (iterator->search_hash) {
 		zend_hash_destroy(iterator->search_hash);
@@ -171,24 +174,22 @@ zend_object* apc_iterator_create(zend_class_entry *ce) {
 static int apc_iterator_search_match(apc_iterator_t *iterator, apc_cache_entry_t *entry) {
 	int rval = 1;
 
-#ifdef ITERATOR_PCRE
 	if (iterator->regex) {
-# if PHP_VERSION_ID >= 70300
+#if PHP_VERSION_ID >= 70300
 		rval = pcre2_match(
 			php_pcre_pce_re(iterator->pce),
-			(PCRE2_SPTR) ZSTR_VAL(entry->key), ZSTR_LEN(entry->key),
+			(PCRE2_SPTR) ZSTR_VAL(&entry->key), ZSTR_LEN(&entry->key),
 			0, 0, iterator->re_match_data, php_pcre_mctx()) >= 0;
-# else
+#else
 		rval = pcre_exec(
 			iterator->pce->re, iterator->pce->extra,
-			ZSTR_VAL(entry->key), ZSTR_LEN(entry->key),
+			ZSTR_VAL(&entry->key), ZSTR_LEN(&entry->key),
 			0, 0, NULL, 0) >= 0;
-# endif
-	}
 #endif
+	}
 
 	if (iterator->search_hash) {
-		rval = zend_hash_exists(iterator->search_hash, entry->key);
+		rval = zend_hash_exists(iterator->search_hash, &entry->key);
 	}
 
 	return rval;
@@ -209,8 +210,9 @@ static int apc_iterator_check_expiry(apc_cache_t* cache, apc_cache_entry_t *entr
 /* }}} */
 
 /* {{{ apc_iterator_fetch_active */
-static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
-	int count = 0;
+static size_t apc_iterator_fetch_active(apc_iterator_t *iterator) {
+	apc_cache_t *cache = apc_user_cache;
+	size_t count = 0;
 	apc_iterator_item_t *item;
 	time_t t = apc_time();
 
@@ -218,12 +220,16 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
 		apc_iterator_item_dtor(apc_stack_pop(iterator->stack));
 	}
 
-	APC_RLOCK(apc_user_cache->header);
+	if (!apc_cache_rlock(cache)) {
+		return count;
+	}
+
 	php_apc_try {
-		while (count <= iterator->chunk_size && iterator->slot_idx < apc_user_cache->nslots) {
-			apc_cache_entry_t *entry = apc_user_cache->slots[iterator->slot_idx];
-			while (entry) {
-				if (apc_iterator_check_expiry(apc_user_cache, entry, t)) {
+		while (count <= iterator->chunk_size && iterator->slot_idx < cache->nslots) {
+			uintptr_t entry_offset = cache->slots[iterator->slot_idx];
+			while (entry_offset) {
+				apc_cache_entry_t *entry = ENTRYAT(entry_offset);
+				if (apc_iterator_check_expiry(cache, entry, t)) {
 					if (apc_iterator_search_match(iterator, entry)) {
 						count++;
 						item = apc_iterator_item_ctor(iterator, entry);
@@ -232,13 +238,13 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
 						}
 					}
 				}
-				entry = entry->next;
+				entry_offset = entry->next;
 			}
 			iterator->slot_idx++;
 		}
 	} php_apc_finally {
 		iterator->stack_idx = 0;
-		APC_RUNLOCK(apc_user_cache->header)
+		apc_cache_runlock(cache);
 	} php_apc_end_try();
 
 	return count;
@@ -246,19 +252,24 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
 /* }}} */
 
 /* {{{ apc_iterator_fetch_deleted */
-static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
-	int count = 0;
+static size_t apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
+	apc_cache_t *cache = apc_user_cache;
+	size_t count = 0;
 	apc_iterator_item_t *item;
 
-	APC_RLOCK(apc_user_cache->header);
+	if (!apc_cache_rlock(cache)) {
+		return count;
+	}
+
 	php_apc_try {
-		apc_cache_entry_t *entry = apc_user_cache->header->gc;
-		while (entry && count <= iterator->slot_idx) {
+		uintptr_t entry_offset = cache->header->gc;
+		while (entry_offset && count <= iterator->slot_idx) {
 			count++;
-			entry = entry->next;
+			entry_offset = ENTRYAT(entry_offset)->next;
 		}
 		count = 0;
-		while (entry && count < iterator->chunk_size) {
+		while (entry_offset && count < iterator->chunk_size) {
+			apc_cache_entry_t *entry = ENTRYAT(entry_offset);
 			if (apc_iterator_search_match(iterator, entry)) {
 				count++;
 				item = apc_iterator_item_ctor(iterator, entry);
@@ -266,12 +277,12 @@ static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
 					apc_stack_push(iterator->stack, item);
 				}
 			}
-			entry = entry->next;
+			entry_offset = entry->next;
 		}
 	} php_apc_finally {
 		iterator->slot_idx += count;
 		iterator->stack_idx = 0;
-		APC_RUNLOCK(apc_user_cache->header);
+		apc_cache_runlock(cache);
 	} php_apc_end_try();
 
 	return count;
@@ -280,44 +291,46 @@ static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
 
 /* {{{ apc_iterator_totals */
 static void apc_iterator_totals(apc_iterator_t *iterator) {
+	apc_cache_t *cache = apc_user_cache;
 	time_t t = apc_time();
-	int i;
 
-	APC_RLOCK(apc_user_cache->header);
+	if (!apc_cache_rlock(cache)) {
+		return;
+	}
+
 	php_apc_try {
-		for (i=0; i < apc_user_cache->nslots; i++) {
-			apc_cache_entry_t *entry = apc_user_cache->slots[i];
-			while (entry) {
-				if (apc_iterator_check_expiry(apc_user_cache, entry, t)) {
+		size_t i;
+
+		for (i=0; i < cache->nslots; i++) {
+			uintptr_t entry_offset = cache->slots[i];
+			while (entry_offset) {
+				apc_cache_entry_t *entry = ENTRYAT(entry_offset);
+				if (apc_iterator_check_expiry(cache, entry, t)) {
 					if (apc_iterator_search_match(iterator, entry)) {
 						iterator->size += entry->mem_size;
 						iterator->hits += entry->nhits;
 						iterator->count++;
 					}
 				}
-				entry = entry->next;
+				entry_offset = entry->next;
 			}
 		}
 	} php_apc_finally {
 		iterator->totals_flag = 1;
-		APC_RUNLOCK(apc_user_cache->header);
+		apc_cache_runlock(cache);
 	} php_apc_end_try();
 }
 /* }}} */
 
-void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long format, zend_long chunk_size, zend_long list)
+void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long format, size_t chunk_size, zend_long list)
 {
 	if (!APCG(enabled)) {
-		apc_error("APC must be enabled to use " APC_ITERATOR_NAME);
-	}
-
-	if (chunk_size < 0) {
-		apc_error(APC_ITERATOR_NAME " chunk size must be 0 or greater");
+		zend_throw_error(NULL, "APC must be enabled to use APCUIterator");
 		return;
 	}
 
 	if (format > APC_ITER_ALL) {
-		apc_error(APC_ITERATOR_NAME " format is invalid");
+		apc_error("APCUIterator format is invalid");
 		return;
 	}
 
@@ -326,7 +339,7 @@ void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long for
 	} else if (list == APC_LIST_DELETED) {
 		iterator->fetch = apc_iterator_fetch_deleted;
 	} else {
-		apc_warning(APC_ITERATOR_NAME " invalid list type");
+		apc_warning("APCUIterator invalid list type");
 		return;
 	}
 
@@ -343,7 +356,6 @@ void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long for
 	iterator->regex = NULL;
 	iterator->search_hash = NULL;
 	if (search && Z_TYPE_P(search) == IS_STRING && Z_STRLEN_P(search)) {
-#ifdef ITERATOR_PCRE
 		iterator->regex = zend_string_copy(Z_STR_P(search));
 		iterator->pce = pcre_get_compiled_regex_cache(iterator->regex);
 
@@ -353,12 +365,9 @@ void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long for
 			iterator->regex = NULL;
 		}
 
-# if PHP_VERSION_ID >= 70300
+#if PHP_VERSION_ID >= 70300
 		iterator->re_match_data = pcre2_match_data_create_from_pattern(
 			php_pcre_pce_re(iterator->pce), php_pcre_gctx());
-# endif
-#else
-		apc_error("Regular expressions support is not enabled, please enable PCRE for " APC_ITERATOR_NAME " regex support.");
 #endif
 	} else if (search && Z_TYPE_P(search) == IS_ARRAY) {
 		iterator->search_hash = apc_flip_hash(Z_ARRVAL_P(search));
@@ -366,52 +375,52 @@ void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long for
 	iterator->initialized = 1;
 }
 
-/* {{{ proto object APCuIterator::__construct([ mixed search [, long format [, long chunk_size [, long list ]]]]) */
-PHP_METHOD(apc_iterator, __construct) {
+PHP_METHOD(APCUIterator, __construct) {
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 	zend_long format = APC_ITER_ALL;
-	zend_long chunk_size=0;
+	zend_long chunk_size = 0;
 	zval *search = NULL;
 	zend_long list = APC_LIST_ACTIVE;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|zlll", &search, &format, &chunk_size, &list) == FAILURE) {
+	ZEND_PARSE_PARAMETERS_START(0, 4)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_ZVAL_EX(search, 1, 0)
+		Z_PARAM_LONG(format)
+		Z_PARAM_LONG(chunk_size)
+		Z_PARAM_LONG(list)
+	ZEND_PARSE_PARAMETERS_END();
+
+	if (chunk_size < 0) {
+		apc_error("APCUIterator chunk size must be 0 or greater");
 		return;
 	}
 
 	apc_iterator_obj_init(iterator, search, format, chunk_size, list);
 }
-/* }}} */
 
-/* {{{ proto APCuIterator::rewind() */
-PHP_METHOD(apc_iterator, rewind) {
+PHP_METHOD(APCUIterator, rewind) {
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
 	if (zend_parse_parameters_none() == FAILURE) {
 		return;
 	}
 
-	if (iterator->initialized == 0) {
-		RETURN_FALSE;
-	}
+	ENSURE_INITIALIZED(iterator);
 
 	iterator->slot_idx = 0;
 	iterator->stack_idx = 0;
 	iterator->key_idx = 0;
 	iterator->fetch(iterator);
 }
-/* }}} */
 
-/* {{{ proto boolean APCuIterator::valid() */
-PHP_METHOD(apc_iterator, valid) {
+PHP_METHOD(APCUIterator, valid) {
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
 	if (zend_parse_parameters_none() == FAILURE) {
 		return;
 	}
 
-	if (iterator->initialized == 0) {
-		RETURN_FALSE;
-	}
+	ENSURE_INITIALIZED(iterator);
 
 	if (apc_stack_size(iterator->stack) == iterator->stack_idx) {
 		iterator->fetch(iterator);
@@ -419,10 +428,8 @@ PHP_METHOD(apc_iterator, valid) {
 
 	RETURN_BOOL(apc_stack_size(iterator->stack) == 0 ? 0 : 1);
 }
-/* }}} */
 
-/* {{{ proto mixed APCuIterator::current() */
-PHP_METHOD(apc_iterator, current) {
+PHP_METHOD(APCUIterator, current) {
 	apc_iterator_item_t *item;
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
@@ -430,23 +437,20 @@ PHP_METHOD(apc_iterator, current) {
 		return;
 	}
 
-	if (iterator->initialized == 0) {
-		RETURN_FALSE;
-	}
+	ENSURE_INITIALIZED(iterator);
 
 	if (apc_stack_size(iterator->stack) == iterator->stack_idx) {
 		if (iterator->fetch(iterator) == 0) {
-			RETURN_FALSE;
+			zend_throw_error(NULL, "Cannot call current() on invalid iterator");
+			return;
 		}
 	}
 
 	item = apc_stack_get(iterator->stack, iterator->stack_idx);
 	ZVAL_COPY(return_value, &item->value);
 }
-/* }}} */
 
-/* {{{ proto string APCuIterator::key() */
-PHP_METHOD(apc_iterator, key) {
+PHP_METHOD(APCUIterator, key) {
 	apc_iterator_item_t *item;
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
@@ -454,13 +458,11 @@ PHP_METHOD(apc_iterator, key) {
 		return;
 	}
 
-	if (iterator->initialized == 0 || apc_stack_size(iterator->stack) == 0) {
-		RETURN_FALSE;
-	}
-
+	ENSURE_INITIALIZED(iterator);
 	if (apc_stack_size(iterator->stack) == iterator->stack_idx) {
 		if (iterator->fetch(iterator) == 0) {
-			RETURN_FALSE;
+			zend_throw_error(NULL, "Cannot call key() on invalid iterator");
+			return;
 		}
 	}
 
@@ -472,38 +474,31 @@ PHP_METHOD(apc_iterator, key) {
 		RETURN_LONG(iterator->key_idx);
 	}
 }
-/* }}} */
 
-/* {{{ proto APCuIterator::next() */
-PHP_METHOD(apc_iterator, next) {
+PHP_METHOD(APCUIterator, next) {
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
 	if (zend_parse_parameters_none() == FAILURE) {
 		return;
 	}
 
-	if (iterator->initialized == 0 || apc_stack_size(iterator->stack) == 0) {
-		RETURN_FALSE;
+	ENSURE_INITIALIZED(iterator);
+	if (apc_stack_size(iterator->stack) == 0) {
+		return;
 	}
 
 	iterator->stack_idx++;
 	iterator->key_idx++;
-
-	RETURN_TRUE;
 }
-/* }}} */
 
-/* {{{ proto long APCuIterator::getTotalHits() */
-PHP_METHOD(apc_iterator, getTotalHits) {
+PHP_METHOD(APCUIterator, getTotalHits) {
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
 	if (zend_parse_parameters_none() == FAILURE) {
 		return;
 	}
 
-	if (iterator->initialized == 0) {
-		RETURN_FALSE;
-	}
+	ENSURE_INITIALIZED(iterator);
 
 	if (iterator->totals_flag == 0) {
 		apc_iterator_totals(iterator);
@@ -513,17 +508,14 @@ PHP_METHOD(apc_iterator, getTotalHits) {
 }
 /* }}} */
 
-/* {{{ proto long APCuIterator::getTotalSize() */
-PHP_METHOD(apc_iterator, getTotalSize) {
+PHP_METHOD(APCUIterator, getTotalSize) {
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
 	if (zend_parse_parameters_none() == FAILURE) {
 		return;
 	}
 
-	if (iterator->initialized == 0) {
-		RETURN_FALSE;
-	}
+	ENSURE_INITIALIZED(iterator);
 
 	if (iterator->totals_flag == 0) {
 		apc_iterator_totals(iterator);
@@ -531,19 +523,15 @@ PHP_METHOD(apc_iterator, getTotalSize) {
 
 	RETURN_LONG(iterator->size);
 }
-/* }}} */
 
-/* {{{ proto long APCuIterator::getTotalCount() */
-PHP_METHOD(apc_iterator, getTotalCount) {
+PHP_METHOD(APCUIterator, getTotalCount) {
 	apc_iterator_t *iterator = apc_iterator_fetch(getThis());
 
 	if (zend_parse_parameters_none() == FAILURE) {
 		return;
 	}
 
-	if (iterator->initialized == 0) {
-		RETURN_FALSE;
-	}
+	ENSURE_INITIALIZED(iterator);
 
 	if (iterator->totals_flag == 0) {
 		apc_iterator_totals(iterator);
@@ -551,40 +539,12 @@ PHP_METHOD(apc_iterator, getTotalCount) {
 
 	RETURN_LONG(iterator->count);
 }
-/* }}} */
-
-/* {{{ arginfo */
-ZEND_BEGIN_ARG_INFO_EX(arginfo_apc_iterator___construct, 0, 0, 0)
-	ZEND_ARG_INFO(0, search)
-	ZEND_ARG_INFO(0, format)
-	ZEND_ARG_INFO(0, chunk_size)
-	ZEND_ARG_INFO(0, list)
-ZEND_END_ARG_INFO()
-
-ZEND_BEGIN_ARG_INFO_EX(arginfo_apc_iterator_void, 0, 0, 0)
-ZEND_END_ARG_INFO()
-/* }}} */
-
-/* {{{ apc_iterator_functions */
-static zend_function_entry apc_iterator_functions[] = {
-	PHP_ME(apc_iterator, __construct, arginfo_apc_iterator___construct, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, rewind, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, current, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, key, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, next, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, valid, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, getTotalHits, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, getTotalSize, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_ME(apc_iterator, getTotalCount, arginfo_apc_iterator_void, ZEND_ACC_PUBLIC)
-	PHP_FE_END
-};
-/* }}} */
 
 /* {{{ apc_iterator_init */
 int apc_iterator_init(int module_number) {
 	zend_class_entry ce;
 
-	INIT_CLASS_ENTRY(ce, APC_ITERATOR_NAME, apc_iterator_functions);
+	INIT_CLASS_ENTRY(ce, "APCUIterator", class_APCUIterator_methods);
 	apc_iterator_ce = zend_register_internal_class(&ce);
 	apc_iterator_ce->create_object = apc_iterator_create;
 	zend_class_implements(apc_iterator_ce, 1, zend_ce_iterator);
@@ -607,7 +567,7 @@ int apc_iterator_init(int module_number) {
 
 	memcpy(&apc_iterator_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
 
-	apc_iterator_object_handlers.clone_obj = apc_iterator_clone;
+	apc_iterator_object_handlers.clone_obj = NULL;
 	apc_iterator_object_handlers.free_obj = apc_iterator_free;
 	apc_iterator_object_handlers.offset = XtOffsetOf(apc_iterator_t, obj);
 
@@ -626,12 +586,13 @@ int apc_iterator_delete(zval *zobj) {
 	apc_iterator_item_t *item;
 
 	if (!ce || !instanceof_function(ce, apc_iterator_ce)) {
-		apc_error("apc_delete object argument must be instance of " APC_ITERATOR_NAME ".");
+		apc_error("apc_delete object argument must be instance of APCUIterator.");
 		return 0;
 	}
 	iterator = apc_iterator_fetch(zobj);
 
 	if (iterator->initialized == 0) {
+		zend_throw_error(NULL, "Trying to use uninitialized APCUIterator");
 		return 0;
 	}
 
