@@ -36,6 +36,7 @@
 #include "apc_lock.h"
 #include "apc_mutex.h"
 #include "apc_strings.h"
+#include "apc_time.h"
 #include "php_globals.h"
 #include "php_ini.h"
 #include "ext/standard/info.h"
@@ -43,7 +44,12 @@
 #include "ext/standard/flock_compat.h"
 #include "ext/standard/md5.h"
 #include "ext/standard/php_var.h"
-#include "apc_arginfo.h"
+#if PHP_VERSION_ID >= 80000
+# include "php_apc_arginfo.h"
+#else
+# include "php_apc_legacy_arginfo.h"
+#endif
+#include "php74_shim.h"
 
 #ifdef HAVE_SYS_FILE_H
 #include <sys/file.h>
@@ -52,25 +58,10 @@
 #include "SAPI.h"
 #include "php_apc.h"
 
-#if HAVE_SIGACTION
+#ifdef HAVE_SIGACTION
 #include "apc_signal.h"
 #endif
 
-/* {{{ PHP_FUNCTION declarations */
-PHP_FUNCTION(apcu_cache_info);
-PHP_FUNCTION(apcu_clear_cache);
-PHP_FUNCTION(apcu_sma_info);
-PHP_FUNCTION(apcu_key_info);
-PHP_FUNCTION(apcu_store);
-PHP_FUNCTION(apcu_fetch);
-PHP_FUNCTION(apcu_delete);
-PHP_FUNCTION(apcu_add);
-PHP_FUNCTION(apcu_inc);
-PHP_FUNCTION(apcu_dec);
-PHP_FUNCTION(apcu_cas);
-PHP_FUNCTION(apcu_exists);
-/* }}} */
-
 /* {{{ ZEND_DECLARE_MODULE_GLOBALS(apcu) */
 ZEND_DECLARE_MODULE_GLOBALS(apcu)
 
@@ -92,31 +83,21 @@ static void php_apc_init_globals(zend_apcu_globals* apcu_globals)
 	apcu_globals->smart = 0;
 	apcu_globals->preload_path = NULL;
 	apcu_globals->coredump_unmap = 0;
-	apcu_globals->use_request_time = 1;
+	apcu_globals->use_request_time = 0;
 	apcu_globals->serializer_name = NULL;
-	apcu_globals->recursion = 0;
+	apcu_globals->entry_level = 0;
 }
 /* }}} */
 
 /* {{{ PHP_INI */
 
-static PHP_INI_MH(OnUpdateShmSegments) /* {{{ */
-{
-#if APC_MMAP
-	if (zend_atoi(new_value->val, new_value->len)!=1) {
-		php_error_docref(NULL, E_WARNING, "apc.shm_segments setting ignored in MMAP mode");
-	}
-	APCG(shm_segments) = 1;
-#else
-	APCG(shm_segments) = zend_atoi(new_value->val, new_value->len);
-#endif
-	return SUCCESS;
-}
-/* }}} */
-
 static PHP_INI_MH(OnUpdateShmSize) /* {{{ */
 {
+#if PHP_VERSION_ID >= 80200
+	zend_long s = zend_ini_parse_quantity_warn(new_value, entry->name);
+#else
 	zend_long s = zend_atol(new_value->val, new_value->len);
+#endif
 
 	if (s <= 0) {
 		return FAILURE;
@@ -135,22 +116,49 @@ static PHP_INI_MH(OnUpdateShmSize) /* {{{ */
 }
 /* }}} */
 
+#if defined(APC_MMAP)
+static PHP_INI_MH(OnUpdateMmapHugepageSize) /* {{{ */
+{
+	zend_long s;
+
+#if PHP_VERSION_ID >= 80200
+	s = zend_ini_parse_quantity_warn(new_value, entry->name);
+#else
+	s = zend_atol(new_value->val, new_value->len);
+#endif
+
+	if (s < 0) {
+		php_error_docref(NULL, E_CORE_ERROR, "apc.mmap_hugepage_size must be a positive integer");
+		return FAILURE;
+	}
+
+	if (s & (s - 1)) {
+		php_error_docref(NULL, E_CORE_ERROR, "apc.mmap_hugepage_size must be a power of 2");
+		return FAILURE;
+	}
+
+	APCG(mmap_hugepage_size) = s;
+	return SUCCESS;
+}
+/* }}} */
+#endif
+
 PHP_INI_BEGIN()
 STD_PHP_INI_BOOLEAN("apc.enabled",      "1",    PHP_INI_SYSTEM, OnUpdateBool,              enabled,          zend_apcu_globals, apcu_globals)
-STD_PHP_INI_ENTRY("apc.shm_segments",   "1",    PHP_INI_SYSTEM, OnUpdateShmSegments,       shm_segments,     zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.shm_size",       "32M",  PHP_INI_SYSTEM, OnUpdateShmSize,           shm_size,         zend_apcu_globals, apcu_globals)
-STD_PHP_INI_ENTRY("apc.entries_hint",   "4096", PHP_INI_SYSTEM, OnUpdateLong,              entries_hint,     zend_apcu_globals, apcu_globals)
+STD_PHP_INI_ENTRY("apc.entries_hint",   "0",    PHP_INI_SYSTEM, OnUpdateLong,              entries_hint,     zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.gc_ttl",         "3600", PHP_INI_SYSTEM, OnUpdateLong,              gc_ttl,           zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.ttl",            "0",    PHP_INI_SYSTEM, OnUpdateLong,              ttl,              zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.smart",          "0",    PHP_INI_SYSTEM, OnUpdateLong,              smart,            zend_apcu_globals, apcu_globals)
-#if APC_MMAP
-STD_PHP_INI_ENTRY("apc.mmap_file_mask",  NULL,  PHP_INI_SYSTEM, OnUpdateString,            mmap_file_mask,   zend_apcu_globals, apcu_globals)
+#ifdef APC_MMAP
+STD_PHP_INI_ENTRY("apc.mmap_file_mask", NULL,   PHP_INI_SYSTEM, OnUpdateString,            mmap_file_mask,    zend_apcu_globals, apcu_globals)
+STD_PHP_INI_ENTRY("apc.mmap_hugepage_size", "0", PHP_INI_SYSTEM, OnUpdateMmapHugepageSize, mmap_hugepage_size, zend_apcu_globals, apcu_globals)
 #endif
 STD_PHP_INI_BOOLEAN("apc.enable_cli",   "0",    PHP_INI_SYSTEM, OnUpdateBool,              enable_cli,       zend_apcu_globals, apcu_globals)
 STD_PHP_INI_BOOLEAN("apc.slam_defense", "0",    PHP_INI_SYSTEM, OnUpdateBool,              slam_defense,     zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.preload_path", (char*)NULL,              PHP_INI_SYSTEM, OnUpdateString,       preload_path,  zend_apcu_globals, apcu_globals)
 STD_PHP_INI_BOOLEAN("apc.coredump_unmap", "0", PHP_INI_SYSTEM, OnUpdateBool, coredump_unmap, zend_apcu_globals, apcu_globals)
-STD_PHP_INI_BOOLEAN("apc.use_request_time", "1", PHP_INI_ALL, OnUpdateBool, use_request_time,  zend_apcu_globals, apcu_globals)
+STD_PHP_INI_BOOLEAN("apc.use_request_time", "0", PHP_INI_ALL, OnUpdateBool, use_request_time,  zend_apcu_globals, apcu_globals)
 STD_PHP_INI_ENTRY("apc.serializer", "php", PHP_INI_SYSTEM, OnUpdateStringUnempty, serializer_name, zend_apcu_globals, apcu_globals)
 PHP_INI_END()
 
@@ -172,7 +180,7 @@ static PHP_MINFO_FUNCTION(apcu)
 #else
 	php_info_print_table_row(2, "APCu Debugging", "Disabled");
 #endif
-#if APC_MMAP
+#ifdef APC_MMAP
 	php_info_print_table_row(2, "MMAP Support", "Enabled");
 	php_info_print_table_row(2, "MMAP File Mask", APCG(mmap_file_mask));
 #else
@@ -239,10 +247,12 @@ static PHP_MINIT_FUNCTION(apcu)
 	if (APCG(enabled)) {
 
 		if (!APCG(initialized)) {
-#if APC_MMAP
-			char *mmap_file_mask = APCG(mmap_file_mask);
-#else
 			char *mmap_file_mask = NULL;
+			zend_long mmap_hugepage_size = 0;
+
+#ifdef APC_MMAP
+			mmap_file_mask = APCG(mmap_file_mask);
+			mmap_hugepage_size = APCG(mmap_hugepage_size);
 #endif
 
 			/* ensure this runs only once */
@@ -251,7 +261,7 @@ static PHP_MINIT_FUNCTION(apcu)
 			/* initialize shared memory allocator */
 			apc_sma_init(
 				&apc_sma, (void **) &apc_user_cache, (apc_sma_expunge_f) apc_cache_default_expunge,
-				APCG(shm_segments), APCG(shm_size), mmap_file_mask);
+				APCG(shm_size), APC_ENTRY_SIZE(0), mmap_file_mask, mmap_hugepage_size);
 
 			REGISTER_LONG_CONSTANT(APC_SERIALIZER_CONSTANT, (zend_long)&_apc_register_serializer, CONST_PERSISTENT | CONST_CS);
 
@@ -304,7 +314,7 @@ static PHP_MSHUTDOWN_FUNCTION(apcu)
 			APCG(initialized) = 0;
 		}
 
-#if HAVE_SIGACTION
+#ifdef HAVE_SIGACTION
 		apc_shutdown_signals();
 #endif
 	}
@@ -329,7 +339,7 @@ static PHP_RINIT_FUNCTION(apcu)
 			apc_cache_serializer(apc_user_cache, APCG(serializer_name));
 		}
 
-#if HAVE_SIGACTION
+#ifdef HAVE_SIGACTION
 		apc_set_signals();
 #endif
 	}
@@ -354,9 +364,10 @@ PHP_FUNCTION(apcu_cache_info)
 {
 	zend_bool limited = 0;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|b", &limited) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(0,1)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_BOOL(limited)
+	ZEND_PARSE_PARAMETERS_END();
 
 	if (!apc_cache_info(return_value, apc_user_cache, limited)) {
 		php_error_docref(NULL, E_WARNING, "No APC info available.  Perhaps APC is not enabled? Check apc.enabled in your ini file");
@@ -370,9 +381,9 @@ PHP_FUNCTION(apcu_key_info)
 {
 	zend_string *key;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S", &key) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 1)
+		Z_PARAM_STR(key)
+	ZEND_PARSE_PARAMETERS_END();
 
 	apc_cache_stat(apc_user_cache, key, return_value);
 } /* }}} */
@@ -380,24 +391,22 @@ PHP_FUNCTION(apcu_key_info)
 /* {{{ proto array apcu_sma_info([bool limited]) */
 PHP_FUNCTION(apcu_sma_info)
 {
-	apc_sma_info_t* info;
-	zval block_lists;
-	int i;
 	zend_bool limited = 0;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|b", &limited) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(0, 1)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_BOOL(limited)
+	ZEND_PARSE_PARAMETERS_END();
 
-	info = apc_sma_info(&apc_sma, limited);
+	apc_sma_info_t *info = apc_sma_info(&apc_sma, limited);
 
 	if (!info) {
 		php_error_docref(NULL, E_WARNING, "No APC SMA info available.  Perhaps APC is disabled via apc.enabled?");
 		RETURN_FALSE;
 	}
-	array_init(return_value);
 
-	add_assoc_long(return_value, "num_seg", info->num_seg);
+	array_init(return_value);
+	add_assoc_long(return_value, "num_seg", 1);
 	add_assoc_double(return_value, "seg_size", (double)info->seg_size);
 	add_assoc_double(return_value, "avail_mem", (double)apc_sma_get_avail_mem(&apc_sma));
 
@@ -406,24 +415,25 @@ PHP_FUNCTION(apcu_sma_info)
 		return;
 	}
 
-	array_init(&block_lists);
+	/* generate list of free blocks */
+	apc_sma_link_t *p;
+	zval list;
+	array_init(&list);
+	for (p = info->list; p != NULL; p = p->next) {
+		zval link;
 
-	for (i = 0; i < info->num_seg; i++) {
-		apc_sma_link_t* p;
-		zval list;
-
-		array_init(&list);
-		for (p = info->list[i]; p != NULL; p = p->next) {
-			zval link;
-
-			array_init(&link);
-
-			add_assoc_long(&link, "size", p->size);
-			add_assoc_long(&link, "offset", p->offset);
-			add_next_index_zval(&list, &link);
-		}
-		add_next_index_zval(&block_lists, &list);
+		array_init(&link);
+		add_assoc_long(&link, "size", p->size);
+		add_assoc_long(&link, "offset", p->offset);
+		add_next_index_zval(&list, &link);
 	}
+
+	/* Since support for multiple shm segments has been dropped, the "block_lists" array
+	 * only exists to ensure compatibility with existing PHP scripts. */
+	zval block_lists;
+	array_init(&block_lists);
+	add_next_index_zval(&block_lists, &list);
+
 	add_assoc_zval(return_value, "block_lists", &block_lists);
 	apc_sma_free_info(&apc_sma, info);
 }
@@ -431,7 +441,7 @@ PHP_FUNCTION(apcu_sma_info)
 
 /* {{{ php_apc_update  */
 zend_bool php_apc_update(
-		zend_string *key, apc_cache_updater_t updater, void *data,
+		zend_string *key, apc_cache_atomic_updater_t updater, void *data,
 		zend_bool insert_if_not_found, time_t ttl)
 {
 	if (APCG(serializer_name)) {
@@ -439,7 +449,7 @@ zend_bool php_apc_update(
 		apc_cache_serializer(apc_user_cache, APCG(serializer_name));
 	}
 
-	return apc_cache_update(apc_user_cache, key, updater, data, insert_if_not_found, ttl);
+	return apc_cache_atomic_update_long(apc_user_cache, key, updater, data, insert_if_not_found, ttl);
 }
 /* }}} */
 
@@ -451,15 +461,19 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 	zval *val = NULL;
 	zend_long ttl = 0L;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z|zl", &key, &val, &ttl) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 3)
+		Z_PARAM_ZVAL(key)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_ZVAL(val)
+		Z_PARAM_LONG(ttl)
+	ZEND_PARSE_PARAMETERS_END();
 
 	if (APCG(serializer_name)) {
 		/* Avoid race conditions between MINIT of apc and serializer exts like igbinary */
 		apc_cache_serializer(apc_user_cache, APCG(serializer_name));
 	}
 
+	/* TODO: Port to array|string for PHP 8? */
 	if (Z_TYPE_P(key) == IS_ARRAY) {
 		zval *hentry;
 		zend_string *hkey;
@@ -474,12 +488,14 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 		ZEND_HASH_FOREACH_KEY_VAL(hash, hkey_idx, hkey, hentry) {
 			ZVAL_DEREF(hentry);
 			if (hkey) {
-				if (!apc_cache_store(apc_user_cache, hkey, hentry, (uint32_t) ttl, exclusive)) {
-					zend_hash_add_new(Z_ARRVAL_P(return_value), hkey, &fail_zv);
-				}
+				zend_string_addref(hkey);
 			} else {
-				zend_hash_index_add_new(Z_ARRVAL_P(return_value), hkey_idx, &fail_zv);
+				hkey = zend_long_to_str(hkey_idx);
 			}
+			if (!apc_cache_store(apc_user_cache, hkey, hentry, (uint32_t) ttl, exclusive)) {
+				zend_symtable_add_new(Z_ARRVAL_P(return_value), hkey, &fail_zv);
+			}
+			zend_string_release(hkey);
 		} ZEND_HASH_FOREACH_END();
 		return;
 	} else if (Z_TYPE_P(key) == IS_STRING) {
@@ -499,6 +515,9 @@ static void apc_store_helper(INTERNAL_FUNCTION_PARAMETERS, const zend_bool exclu
 /* {{{ proto bool apcu_enabled(void)
 	returns true when apcu is usable in the current environment */
 PHP_FUNCTION(apcu_enabled) {
+	if (zend_parse_parameters_none() == FAILURE) {
+		return;
+	}
 	RETURN_BOOL(APCG(enabled));
 }
 /* }}} */
@@ -520,20 +539,14 @@ PHP_FUNCTION(apcu_add) {
 /* {{{ php_inc_updater */
 
 struct php_inc_updater_args {
-	zval step;
-	zval rval;
+	zend_long step;
+	zend_long rval;
 };
 
-static zend_bool php_inc_updater(apc_cache_t* cache, apc_cache_entry_t* entry, void* data) {
-	struct php_inc_updater_args *args = (struct php_inc_updater_args*) data;
-
-	if (Z_TYPE(entry->val) == IS_LONG) {
-		fast_long_add_function(&entry->val, &entry->val, &args->step);
-		ZVAL_COPY_VALUE(&args->rval, &entry->val);
-		return 1;
-	}
-
-	return 0;
+static zend_bool php_inc_updater(apc_cache_t *cache, zend_long *entry, void *data) {
+	struct php_inc_updater_args *args = (struct php_inc_updater_args *) data;
+	args->rval = ATOMIC_ADD(*entry, args->step);
+	return 1;
 }
 
 /* {{{ proto long apcu_inc(string key [, long step [, bool& success [, long ttl]]])
@@ -544,26 +557,24 @@ PHP_FUNCTION(apcu_inc) {
 	zend_long step = 1, ttl = 0;
 	zval *success = NULL;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|lzl", &key, &step, &success, &ttl) == FAILURE) {
-		return;
-	}
-
-	if (success) {
-		ZVAL_DEREF(success);
-		zval_ptr_dtor(success);
-	}
-
-	ZVAL_LONG(&args.step, step);
+	ZEND_PARSE_PARAMETERS_START(1, 4)
+		Z_PARAM_STR(key)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_LONG(step)
+		Z_PARAM_ZVAL(success)
+		Z_PARAM_LONG(ttl)
+	ZEND_PARSE_PARAMETERS_END();
 
+	args.step = step;
 	if (php_apc_update(key, php_inc_updater, &args, 1, ttl)) {
 		if (success) {
-			ZVAL_TRUE(success);
+			ZEND_TRY_ASSIGN_REF_TRUE(success);
 		}
-		RETURN_ZVAL(&args.rval, 0, 0);
+		RETURN_LONG(args.rval);
 	}
 
 	if (success) {
-		ZVAL_FALSE(success);
+		ZEND_TRY_ASSIGN_REF_FALSE(success);
 	}
 
 	RETURN_FALSE;
@@ -578,27 +589,25 @@ PHP_FUNCTION(apcu_dec) {
 	zend_long step = 1, ttl = 0;
 	zval *success = NULL;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "S|lzl", &key, &step, &success, &ttl) == FAILURE) {
-		return;
-	}
-
-	if (success) {
-		ZVAL_DEREF(success);
-		zval_ptr_dtor(success);
-	}
-
-	ZVAL_LONG(&args.step, 0 - step);
+	ZEND_PARSE_PARAMETERS_START(1, 4)
+		Z_PARAM_STR(key)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_LONG(step)
+		Z_PARAM_ZVAL(success)
+		Z_PARAM_LONG(ttl)
+	ZEND_PARSE_PARAMETERS_END();
 
+	args.step = 0 - step;
 	if (php_apc_update(key, php_inc_updater, &args, 1, ttl)) {
 		if (success) {
-			ZVAL_TRUE(success);
+			ZEND_TRY_ASSIGN_REF_TRUE(success);
 		}
 
-		RETURN_ZVAL(&args.rval, 0, 0);
+		RETURN_LONG(args.rval);
 	}
 
 	if (success) {
-		ZVAL_FALSE(success);
+		ZEND_TRY_ASSIGN_REF_FALSE(success);
 	}
 
 	RETURN_FALSE;
@@ -620,9 +629,11 @@ PHP_FUNCTION(apcu_cas) {
 	zend_string *key;
 	zend_long vals[2];
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "Sll", &key, &vals[0], &vals[1]) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(3, 3)
+		Z_PARAM_STR(key)
+		Z_PARAM_LONG(vals[0])
+		Z_PARAM_LONG(vals[1])
+	ZEND_PARSE_PARAMETERS_END();
 
 	if (APCG(serializer_name)) {
 		/* Avoid race conditions between MINIT of apc and serializer exts like igbinary */
@@ -639,29 +650,23 @@ PHP_FUNCTION(apcu_fetch) {
 	zval *key;
 	zval *success = NULL;
 	time_t t;
+	int result;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z|z", &key, &success) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 2)
+		Z_PARAM_ZVAL(key)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_ZVAL(success)
+	ZEND_PARSE_PARAMETERS_END();
 
 	t = apc_time();
 
-	if (success) {
-		ZVAL_DEREF(success);
-		zval_ptr_dtor(success);
-		ZVAL_FALSE(success);
-	}
-
 	if (Z_TYPE_P(key) != IS_STRING && Z_TYPE_P(key) != IS_ARRAY) {
 		convert_to_string(key);
 	}
 
+	/* TODO: Port to array|string for PHP 8? */
 	if (Z_TYPE_P(key) == IS_STRING) {
-		if (apc_cache_fetch(apc_user_cache, Z_STR_P(key), t, return_value)) {
-			if (success) {
-				ZVAL_TRUE(success);
-			}
-		} else { RETVAL_FALSE; }
+		result = apc_cache_fetch(apc_user_cache, Z_STR_P(key), t, return_value);
 	} else if (Z_TYPE_P(key) == IS_ARRAY) {
 		zval *hentry;
 
@@ -679,12 +684,16 @@ PHP_FUNCTION(apcu_fetch) {
 				apc_warning("apc_fetch() expects a string or array of strings.");
 			}
 		} ZEND_HASH_FOREACH_END();
-
-		if (success) {
-			ZVAL_TRUE(success);
-		}
+		result = 1;
 	} else {
 		apc_warning("apc_fetch() expects a string or array of strings.");
+		result = 0;
+	}
+
+	if (success) {
+		ZEND_TRY_ASSIGN_REF_BOOL(success, result);
+	}
+	if (!result) {
 		RETURN_FALSE;
 	}
 }
@@ -696,9 +705,9 @@ PHP_FUNCTION(apcu_exists) {
 	zval *key;
 	time_t t;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z", &key) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 1)
+		Z_PARAM_ZVAL(key)
+	ZEND_PARSE_PARAMETERS_END();
 
 	t = apc_time();
 
@@ -706,6 +715,7 @@ PHP_FUNCTION(apcu_exists) {
 		convert_to_string(key);
 	}
 
+	/* TODO: Port to array|string for PHP 8? */
 	if (Z_TYPE_P(key) == IS_STRING) {
 		RETURN_BOOL(apc_cache_exists(apc_user_cache, Z_STR_P(key), t));
 	} else if (Z_TYPE_P(key) == IS_ARRAY) {
@@ -737,9 +747,9 @@ PHP_FUNCTION(apcu_exists) {
 PHP_FUNCTION(apcu_delete) {
 	zval *keys;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "z", &keys) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 1)
+		Z_PARAM_ZVAL(keys)
+	ZEND_PARSE_PARAMETERS_END();
 
 	if (Z_TYPE_P(keys) == IS_STRING) {
 		RETURN_BOOL(apc_cache_delete(apc_user_cache, Z_STR_P(keys)));
@@ -767,15 +777,18 @@ PHP_FUNCTION(apcu_delete) {
 }
 
 PHP_FUNCTION(apcu_entry) {
-	zval *key = NULL;
+	zend_string *key;
 	zend_fcall_info fci = empty_fcall_info;
 	zend_fcall_info_cache fcc = empty_fcall_info_cache;
 	zend_long ttl = 0L;
 	zend_long now = apc_time();
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "zf|l", &key, &fci, &fcc, &ttl) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(2, 3)
+		Z_PARAM_STR(key)
+		Z_PARAM_FUNC(fci, fcc)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_LONG(ttl)
+	ZEND_PARSE_PARAMETERS_END();
 
 	apc_cache_entry(apc_user_cache, key, &fci, &fcc, ttl, now, return_value);
 }
@@ -786,9 +799,10 @@ PHP_FUNCTION(apcu_entry) {
 PHP_FUNCTION(apcu_inc_request_time) {
 	zend_long by = 1;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "|l", &by) != SUCCESS) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(0, 1)
+		Z_PARAM_OPTIONAL
+		Z_PARAM_LONG(by)
+	ZEND_PARSE_PARAMETERS_END();
 
 	if (!APCG(use_request_time)) {
 		php_error_docref(NULL, E_WARNING,
@@ -803,35 +817,12 @@ PHP_FUNCTION(apcu_inc_request_time) {
 }
 #endif
 
-/* {{{ apcu_functions[] */
-zend_function_entry apcu_functions[] = {
-	PHP_FE(apcu_cache_info,         arginfo_apcu_cache_info)
-	PHP_FE(apcu_clear_cache,        arginfo_apcu_clear_cache)
-	PHP_FE(apcu_sma_info,           arginfo_apcu_sma_info)
-	PHP_FE(apcu_key_info,           arginfo_apcu_key_info)
-	PHP_FE(apcu_enabled,            arginfo_apcu_enabled)
-	PHP_FE(apcu_store,              arginfo_apcu_store)
-	PHP_FE(apcu_fetch,              arginfo_apcu_fetch)
-	PHP_FE(apcu_delete,             arginfo_apcu_delete)
-	PHP_FE(apcu_add,                arginfo_apcu_store)
-	PHP_FE(apcu_inc,                arginfo_apcu_inc)
-	PHP_FE(apcu_dec,                arginfo_apcu_inc)
-	PHP_FE(apcu_cas,                arginfo_apcu_cas)
-	PHP_FE(apcu_exists,             arginfo_apcu_exists)
-	PHP_FE(apcu_entry,				arginfo_apcu_entry)
-#ifdef APC_DEBUG
-	PHP_FE(apcu_inc_request_time,   NULL)
-#endif
-	PHP_FE_END
-};
-/* }}} */
-
 /* {{{ module definition structure */
 
 zend_module_entry apcu_module_entry = {
 	STANDARD_MODULE_HEADER,
 	PHP_APCU_EXTNAME,
-	apcu_functions,
+	ext_functions,
 	PHP_MINIT(apcu),
 	PHP_MSHUTDOWN(apcu),
 	PHP_RINIT(apcu),
